{-# OPTIONS_GHC -fglasgow-exts #-}
module AdvancedMatch where

import System.IO.SSC.Haskell98 as SSC
import Control.Exception
import Data.Dynamic
import Control.Monad
import Control.Monad.Reader
import System.Random
import Data.List
import Data.IORef

import Match
import Game

nwins :: Int -> Character -> Character -> IO Int
nwins n x y = liftM (length . filter (==Winner First)) $ replicateM n $ quietMatch x y

nwinsG :: RndGen -> Int -> Character -> Character -> IO Int
nwinsG rg n x y = liftM (length . filter (==Winner First)) $ mapM f $ take n gens
  where
    gens = unfoldr (Just . split) rg
    f gen = liftM fst $ matchEx opt x y
      where
        opt = defaultOption{ optGen = Just gen }

winningReplayC :: Character -> Character -> IO ()
winningReplayC c0 c1 = loop 100 c0 c1
  where
    loop 0 _ _ = SSC.putStrLn "100回試したけど勝てなかった"
    loop n c0 c1 = do
      (r, g) <- newMatch (return ()) False c0 c1
      if r == Winner First
        then do
          replayMatch g c0 c1
          return ()
        else loop (n-1) c0 c1

newtype InfoSteal = InfoSteal GameInfo
  deriving (Typeable)

steals :: Exception -> Maybe GameInfo
steals e = do
  d <- dynExceptions e
  InfoSteal gi <- fromDynamic d
  return gi

getInfo :: Option -> Character -> Character -> Game Bool -> IO GameInfo
getInfo opt c0 c1 stop = do
  r <- newIORef False
  catchJust steals (loop r) return
  where
    loop r = do
      (_, gi) <- matchEx opt{ optHook = hook r } c0 c1
      return gi
    hook r = do
      v <- liftIO $ readIORef r
      when (not v) $ do
        p <- stop
        when p $ do
          liftIO $ writeIORef r True
          gi <- ask
          liftIO $ throwIO $ DynException $ toDyn $ InfoSteal gi
