module Skills
  ( fireCracker
  , rest
  , pressurize
  , avenger
  , slash
  , irruption
  , flameBind
  , heatWave
  , fieryCloak
  , kusanagiNoKen
  , aquaForm
  , shieldMist
  , restore
  , poisonousFang
  , whirlwind
  , skillEnsnare
  , nightShrieker
  , drainLife
  , iaTemuNoNoroiNoKen
  , sarasouju
  , paradiseLost
  , oruganrode
  , resutoroose
  , deliriumClaw
  , angelFire
  , reevatein
  , flameDevourer
  , sharumakihyuNoTousi
  , stormBringer
  , iwaugachiNoKen
  , weatheringBlade
  , flashFire
  , reflection
  , silence
  , kyutosuNoYoumaku
  , cutthroat
  ) where

import Game
import Run
import CommonOperations
import GameUtils

import Control.Monad
import Data.List
import Data.Maybe
import qualified Data.Set as S
import qualified Data.Map as M

defaultSkill :: Skill
defaultSkill = Sk
  { skName = "default skill name"
  , skCosts = [error "default skill cost"]
  , skEffect = Just $ error "default skill effect"
  , skCondition = return UseByDefault
  , skReaction = Nothing
  , skReactionCondition = const $ return UseByDefault
  , skReactionAgainstSkill = Nothing
  , skReactionConditionAgainstSkill = const $ return UseByDefault
  , skAttrs = S.empty
  , skTargets = Nothing
  }

fireCracker = defaultSkill
  { skName = "炸撃"
  , skCosts = [Fire]
  , skEffect = Just $ \inv -> do
      dmg <- liftM (1+) $ rnd 2
      attack_ (opp inv) dmg inv
  , skAttrs = S.fromList [AttackingS]
  }
rest = defaultSkill
  { skName = "安らぎ"
  , skCosts = [Water]
  , skEffect = Just $ \inv -> tryRecover (self inv) 2 inv
  , skAttrs = S.fromList [RecoveringS]
  }
pressurize = defaultSkill
  { skName = "空圧"
  , skCosts = [Wind]
  , skEffect = Just $ \inv -> do
      z <- chooseTargetPanel inv (not . paUsed)
      whenJust z $ \(m, n) -> makeUsed (opp inv) m n
  , skAttrs = S.fromList [ObstructiveS]
  , skTargets = Just $ \cd -> panelTargets (opponent cd) (not . paUsed)
  }
avenger = defaultSkill
  { skName = "報復"
  , skCosts = [Earth]
  , skEffect = Nothing
  , skReaction = Just $ \inv ndmg _ -> return $ let
        atk int = attackInt_ (opp inv) int 1 inv
      in (block 3 ndmg, False, atk)
  , skReactionCondition = return . useIf . atkIsSword
  , skAttrs = S.fromList [ReactiveS, DefensiveS, CounterS]
  }
slash = defaultSkill
  { skName = "斬撃"
  , skCosts = [Sword]
  , skEffect = Just $ \inv -> attack_ (opp inv) 3 inv
  , skAttrs = S.fromList [AttackingS]
  }
irruption = defaultSkill
  { skName = "爆撃"
  , skCosts = [Fire, Fire]
  , skEffect = Just $ \inv -> attack_ (opp inv) 4 inv
  , skAttrs = S.fromList [AttackingS]
  }
flameBind = defaultSkill
  { skName = "火炎縛"
  , skCosts = [Fire, Water]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 2 inv $ \_ -> whenLucky 2 3 $ do
        tgt <- chooseTargetSkill inv skIsSword
        whenJust tgt $ \idx -> putSkillBind (opp inv) idx
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> skillTargets (opponent cd) (skIsSword . ssSkill)
  }
heatWave = defaultSkill
  { skName = "灼熱波"
  , skCosts = [Fire, Wind]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 2 inv $ \_ -> whenLucky 2 3 $ do
        tgt <- chooseTargetPlan inv
        whenJust tgt $ \idx -> putPlanBind (opp inv) idx
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> planTargets (opponent cd) (const True)
  }
fieryCloak = defaultSkill
  { skName = "焦熱の外套"
  , skCosts = [Fire, Earth]
  , skEffect = Just $ \inv -> register (self inv) inv >> return ()
  , skAttrs = S.fromList [ContinuousS, DefensiveS, CounterS]
  }
  where
    register cd inv = hold cd $ \un -> defaultHS
      { hsSource = inv
      , hsReaction = \n atk -> react cd inv un n (invSkill $ atkInv atk)
      , hsInterfereCondition = \atk -> let
            sk = invSkill $ atkInv atk
          in skIsFire sk || invControler (atkInv atk) == opponent cd && skIsSword sk
      , hsUnique = un
      }
    react cd inv un ndmg dsource = return (ndmg', act)
      where
        ndmg'
          | skIsFire dsource = block 1 ndmg
          | otherwise = ndmg
        act int = when (skIsSword dsource) $ do
          mesC cd $ "は焦熱の外套の効果によって反撃した"
          attackInt (opponent cd) int 3 inv (unregister cd un) (const $ return ())
          return ()

kusanagiNoKen = defaultSkill
  { skName = "草薙の剣"
  , skCosts = [Fire, Sword]
  , skEffect = Nothing
  , skReaction = Just reaction
  , skReactionCondition = return . useIf . reactCond
  , skReactionAgainstSkill = Just $ \selfi inv ->
      if skIsAttacking (invSkill inv) -- TODO: should be invIsAttacking?
        then do
          registerReaction inv reaction selfi
          return False
        else return True
  , skReactionConditionAgainstSkill = return . useIf . skIsFire . invSkill
  , skAttrs = S.fromList [ReactiveS, DefensiveS, CounterS]
  }
  where
    reactCond atk = atkIsFire atk && invIsReaction (atkInv atk)
    reaction inv ndmg _ = return (div ndmg 2, True, atk)
      where
        atk int = attackInt_ (opp inv) int 2 inv
aquaForm = defaultSkill
  { skName = "水変化"
  , skCosts = [Water, Water]
  , skEffect = Nothing
  , skReaction = Just $ \inv _ _ -> 
      return (0, False, const $ tryRecover (self inv) 2 inv)
  , skReactionCondition = return . useIf . atkIsSword
  , skAttrs = S.fromList [ReactiveS, DefensiveS, RecoveringS]
  }
shieldMist = defaultSkill
  { skName = "霧の防壁"
  , skCosts = [Water, Wind]
  , skReaction = Just $ \inv ndmg _ -> do
      register inv
      (rdmg, act) <- react ndmg
      return (rdmg, False, act)
  , skEffect = Just $ \inv -> register inv >> return ()
  , skCondition = return DontUseByDefault
  , skAttrs = S.fromList [ReactiveS, DefensiveS, ContinuousS]
  }
  where
    register inv = hold (self inv) $ \un -> defaultHS
      { hsSource = inv
      , hsReaction = \n _ -> react n 
      , hsInterfereCondition = const True
      , hsUnique = un
      }
    react ndmg = return (div ndmg 2, \_ -> return ())

data RestoreTarget = PanelConfuse Int | PanelPoison Int | PanelBind Int | SkillBind Int | SkillWeaken Int | PlanBind Int
  deriving (Eq)

restorePanelTargets :: CharDescr -> (Panel -> Int -> Bool) -> Game [RestoreTarget]
restorePanelTargets cd prop = do
  panels <- getPanels cd
  return $ concatMap f panels
  where
    f pa
      | prop pa n = 
        replicate (paPoisoned pa) (PanelPoison n) ++
          replicate (paConfused pa) (PanelConfuse n) ++
          replicate (paBound pa) (PanelBind n)
      | otherwise = []
      where
        n = paIndex pa

restoreSkillTargets :: CharDescr -> (Skill -> Int -> Game Bool) -> Game [RestoreTarget]
restoreSkillTargets cd prop = do
  sss <- getSSs cd
  liftM concat $ mapM f sss
  where
    f SS{ ssBound = bound, ssWeakened = weakened, ssSkill = skill, ssIndex = n } = do
      ok <- prop skill n
      return $ if ok
        then replicate bound (SkillBind n) ++ replicate weakened (SkillWeaken n)
        else []

restorePlanTargets :: CharDescr -> (Plan -> Int -> Bool) -> Game [RestoreTarget]
restorePlanTargets cd prop = do
  plans <- getPresentPlans cd
  return $ concatMap f plans
  where
    f PS{ psPlan = plan, psBound = bound, psIndex = n }
      | prop plan n = replicate bound (PlanBind n)
      | otherwise = []

restorePlannedTargets :: Invocation -> Game [RestoreTarget]
restorePlannedTargets inv = withCurrentSP cd [] $ \csp -> do
  tgts <- targets inv
  liftM concat $ mapM (f csp) $ S.toList tgts
  where
    cd = self inv
    f csp (PanelTarget cd1 m n)
      | cd1 == cd && m == csp = restorePanelTargets cd (\_ x -> x == n)
    f _ (SkillTarget cd1 m)
      | cd1 == cd = restoreSkillTargets cd (\_ n -> return $ n == m)
    f _ (PlanTarget cd1 m)
      | cd1 == cd = restorePlanTargets cd (\_ n -> n == m)
    f _ _ = return []

restoreAllTargets :: CharDescr -> Game [RestoreTarget]
restoreAllTargets cd = do
  plans <- restorePlanTargets cd (\_ _ -> True)
  skills <- restoreSkillTargets cd (\_ _ -> return True)
  panels <- restorePanelTargets cd (\_ _ -> True)
  return $ plans ++ skills ++ panels

chooseRestoreTarget :: Invocation -> Game (Maybe RestoreTarget)
chooseRestoreTarget inv = do
  tgts <- restorePlannedTargets inv
  tgts' <- if null tgts then restoreAllTargets cd else return tgts
  chooseMaybe tgts'
  where
    cd = invControler inv

restore = defaultSkill
  { skName = "修復"
  , skCosts = [Water, Earth]
  , skEffect = Just $ \inv -> do
      let cd = invControler inv
      tryRecover cd 5 inv
      tgt <- chooseRestoreTarget inv
      case tgt of
        Just (PanelConfuse n) -> do
          m <- getCurrentSP' cd
          cureConfuse cd m n >>= performDelays
        Just (PanelPoison n) -> do
          m <- getCurrentSP' cd
          mesPa cd m n $ "のPoisonが回復した"
          modifyPanel cd m n [AddPoison (-1)]
        Just (PanelBind n) -> do
          m <- getCurrentSP' cd
          curePanelBind cd m n
        Just (SkillBind i) -> cureSkillBind cd i
        Just (SkillWeaken i) -> do
          mesC cd $ "のスキル" ++ show i ++ "のweakenが回復した"
          addSkillWeaken cd i (-1)
        Just (PlanBind i) -> curePlanBind cd i
        Nothing -> return ()
  , skAttrs = S.fromList [RecoveringS]
  , skTargets = Just $ \cd -> do
      mcsp <- getCurrentSP cd
      ats <- restoreAllTargets cd
      return $ S.fromList $ mapMaybe (rt_to_st cd mcsp) ats
  }
  where
    rt_to_st cd mcsp rt = case rt of
      PanelConfuse n -> pt n
      PanelPoison n -> pt n
      PanelBind n -> pt n
      SkillBind n -> Just $ SkillTarget cd n
      SkillWeaken n -> Just $ SkillTarget cd n
      PlanBind n -> Just $ PlanTarget cd n
      where
        pt n = case mcsp of
          Nothing -> Nothing
          Just csp -> Just $ PanelTarget cd csp n
poisonousFang = defaultSkill
  { skName = "毒牙"
  , skCosts = [Water, Sword]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 3 inv $ \_ -> whenLucky 2 3 $ do
        tgt <- chooseTargetPanel inv (const True)
        whenJust tgt $ \(m, n) -> putPoison (opp inv) m n
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> panelTargets (opponent cd) (const True)
  }
whirlwind = defaultSkill
  { skName = "旋風"
  , skCosts = [Wind, Wind]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 1 inv $ \_ -> do
        tgt <- chooseTargetPanel inv (const True)
        whenJust tgt $ \(m, n) -> rollPanelBase (opp inv) m n >>= performDelays
  , skAttrs = S.fromList [AttackingS, ObstructiveS]
  , skTargets = Just $ \cd -> panelTargets (opponent cd) (const True)
  }
skillEnsnare = defaultSkill
  { skName = "陥穽"
  , skCosts = [Wind, Earth]
  , skEffect = Nothing
  , skReaction = Nothing
  , skReactionAgainstSkill = Just $ \_ Inv{ invControler = ocd, invCosts = dcosts } -> do
      mapM_ (ensnare ocd) dcosts
      return False
  , skAttrs = S.fromList [ReactiveS, ObstructiveS]
  }
  where
    ensnare cd (m, n) = do
      mesPa cd m n $ "はこのターン回復しない"
      modifyPanel cd m n [SetEnsnared True]
nightShrieker = defaultSkill
  { skName = "夜鳴き刃"
  , skCosts = [Wind, Sword]
  , skEffect = Just $ \inv -> 
      attackSF (opp inv) 4 inv $ \_ -> whenLucky 2 3 $ do
        tgt <- chooseTargetPanel inv (const True)
        whenJust tgt $ \(m, n) -> putConfuse (opp inv) m n
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> panelTargets (opponent cd) (const True)
  }
drainLife = defaultSkill
  { skName = "生命吸収"
  , skCosts = [Earth, Earth]
  , skEffect = Just $ \inv -> do
      myVit <- getVit $ self inv
      oppVit <- getVit $ opp inv
      let x = max 1 $ myVit - oppVit
      attackSF (opp inv) x inv $ \dmg -> tryRecover (self inv) dmg inv
  , skAttrs = S.fromList [AttackingS, RecoveringS]
  }
iaTemuNoNoroiNoKen = defaultSkill
  { skName = "イア＝テムの呪いの剣"
  , skCosts = [Earth, Sword]
  , skEffect = Just $ \inv -> do
      attackSF (opp inv) 5 inv $ \_ -> whenLucky 1 3 $ do
        tgt <- chooseTargetSkill inv skIsMagic
        whenJust tgt $ \idx -> putSkillBind (opp inv) idx
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> skillTargets (opponent cd) (skIsMagic . ssSkill)
  }
sarasouju = defaultSkill
  { skName = "沙羅双樹"
  , skCosts = [Sword, Sword]
  , skEffect = Just $ \inv -> register (self inv) inv >> return ()
  , skAttrs = S.fromList [ContinuousS, CounterS]
  }
  where
    register cd inv = hold cd $ \un -> defaultHS
      { hsStop = stop
      , hsStart = start
      , hsSource = inv
      , hsReaction = \ndmg _ -> react ndmg inv un
      , hsInterfereCondition = skIsSword . invSkill . atkInv
      , hsUnique = un
      }
      where
        start = modifySarasouju cd (1+)
        stop = do
          modifySarasouju cd (subtract 1)
          sara <- getSarasouju cd
          when (sara == 0) $ do
            hp <- getHp cd
            when (hp <= 0) $ lose cd
        react ndmg inv un = return (ndmg, ra)
          where
            ra int = do
              mesC cd $ "は沙羅双樹の効果によって反撃した"
              atk
              mesC cd $ "は沙羅双樹の効果によって再び反撃した"
              atk
              return ()
              where
                atk = attackInt (opponent cd) int 4 inv (unregister cd un) (const $ return ())
paradiseLost = defaultSkill
  { skName = "崩れゆく聖域"
  , skCosts = [Fire, Earth]
  , skEffect = Just $ \inv -> register (self inv) inv >> return ()
  , skAttrs = S.fromList [ObstructiveS, ContinuousS]
  }
  where
    register cd inv = hold cd $ \un -> defaultHS
      { hsStop = addParadiseLost (-1)
      , hsStart = addParadiseLost 1
      , hsSource = inv
      , hsUnique = un
      }
oruganrode = defaultSkill
  { skName = "オルゴーの滅びの呪文"
  , skCosts = [Fire, Water, Wind]
  , skEffect = Just $ \inv -> register (self inv) inv >> return ()
  , skAttrs = S.fromList [AttackingS, ContinuousS]
  }
  where
    register cd inv = hold cd $ \un -> defaultHS
      { hsFinish = do
          (inv1, neg) <- releaseInv
          when (not neg) $ do
            n <- countContinues
            attack_ (opponent cd) (2^n) inv1
      , hsSource = inv
      , hsUnique = un
      }
      where
        countContinues = liftM (M.size . M.filter ((==inv) . contInv)) $ getCalendar
        releaseInv = do
          tt <- tick
          invokeWithCosts cd (invSlot inv) oruganrode (ReleaseI inv) tt (invCosts inv)
resutoroose = defaultSkill
  { skName = "楽園の抱擁"
  , skCosts = [Fire, Water, Earth]
  , skEffect = Just $ \inv -> do
      tryRecover (self inv) 10 inv
      ch <- char (self inv)
      restoreEverything ch (self inv) inv >>= performDelays
  , skAttrs = S.fromList [RecoveringS]
  }
  where
    restoreEverything ch cd inv = liftM concat $ sequence
      [ restoreParams
      , restorePatterns
      , restorePanels
      , restoreSlots
      , restorePlans
      ]
      where
        restoreParams = do
          fixVit
          pfix "神経反射" chReflex getRefl setRefl
          pfix "知力" chInt getInt setInt
          return []
        fixVit = do
          base <- getBaseVit cd
          when (base /= orig) $ do
            setBaseVit cd orig
            vit <- getVit cd
            let newvit = orig - base + vit
            setVit cd newvit
            mesC cd $ "の体力が" ++ show newvit ++ "に戻った"
          where
            orig = chVit ch
        pfix name chp getp setp = do
          v <- getp cd
          when (v /= orig) $ do
            setp cd orig
            mesC cd $ "の" ++ name ++ "が" ++ show orig ++ "に戻った"
          where
            orig = chp ch
        foreach_missing ch_c getter fidx f = do
          let orig = [0 .. length (ch_c ch) - 1]
          old <- liftM (map fidx) $ getter cd
          let to_restore = orig \\ old
          mapM_ f to_restore
        restorePatterns = do
          foreach_missing chSkillPats getSPSs spsIndex $ restoreSP cd
          return []
        restorePanels = do
          spss <- getSPSs cd
          concatMapM on_pat spss
        on_pat sps = do
          mapM_ (restorePanel cd idx) to_restore
          concatMapM (on_panel idx) $ spsPanels sps
          where
            to_restore = origpanels \\ oldpanels
            origpanels = [0 .. length (chSkillPats ch !! idx) - 1]
            oldpanels = map paIndex $ spsPanels sps
            idx = spsIndex sps
        on_panel m pa = do
          p <- fix_p paPoisoned "Poison" SetPoison
          c <- fix_p paConfused "Confuse" SetConfuse
          b <- fix_p paBound "Bind" SetBind
          r <- whenl (paBaseKind pa /= orig) $ return [SetBaseKind orig]
          (k, ds) <- if (paKind pa == orig) then return ([], []) else do
            mesPa cd m n $ "が" ++ showKind orig ++ "に戻った"
            ds <- notifyPanelChange cd DropByRoll m n
            return ([SetKind orig], ds)
          modifyPanel cd m n $ p ++ c ++ b ++ r ++ k
          return ds
          where
            n = paIndex pa
            orig = paOriginalKind pa
            fix_p pa_n name setter = whenl (pa_n pa /= 0) $ do
              mesPa cd m n $ "の" ++ name ++ "が" ++ sall (pa_n pa) ++ "回復した"
              return [setter 0]
        restoreSlots = do
          foreach_missing chSkills getSSs ssIndex $ restoreSlot cd
          sss <- getSSs cd
          flip mapM_ sss $ \ss -> do
            when (ssBound ss /= 0) $ do
              mesC cd $ "のスキル" ++ show (ssIndex ss) ++ "のBindが" ++ sall (ssBound ss) ++ "回復した"
              addSkillBind cd (ssIndex ss) (-ssBound ss)
            when (ssWeakened ss /= 0) $ do
              mesC cd $ "のスキル" ++ show (ssIndex ss) ++ "のweakenが" ++ sall (ssWeakened ss) ++ "回復した"
              addSkillWeaken cd (ssIndex ss) (-ssWeakened ss)
          return []
        restorePlans = do
          foreach_missing chPlans getPresentPlans psIndex $ restorePlan cd
          plans <- getPresentPlans cd
          flip mapM_ plans $ \ps -> when (psBound ps /= 0) $ do
            mesC cd $ "のプラン" ++ show (psIndex ps) ++ "のBindが" ++ sall (psBound ps) ++ "回復した"
            addPlanBind cd (psIndex ps) (-psBound ps)
          return []
        sall 1 = ""
        sall _ = "全て"
        whenl False _ = return []
        whenl True x = x
deliriumClaw = defaultSkill
  { skName = "狂乱爪"
  , skCosts = [Fire, Water, Sword]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 4 inv $ \_ -> do
        tgt <- chooseTargetPanel inv (const True)
        whenJust tgt $ \(m, n) -> do
          whenLucky 2 3 $ putPoison (opp inv) m n
          whenLucky 2 3 $ putConfuse (opp inv) m n
          whenLucky 2 3 $ putPanelBind (opp inv) m n
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> panelTargets (opponent cd) (const True)
  }
angelFire = defaultSkill
  { skName = "炎天使"
  , skCosts = [Fire, Wind, Earth]
  , skEffect = Just $ \inv -> register (self inv) inv >> return ()
  , skAttrs = S.fromList [ContinuousS, DefensiveS, StrengtheningS]
  }
  where
    register cd inv = hold cd $ \un -> defaultHS
      { hsStop = stop
      , hsStart = start
      , hsSource = inv
      , hsReaction = \ndmg src -> return (block 2 ndmg, const $ return ())
      , hsInterfereCondition = skIsSword . invSkill . atkInv
      , hsUnique = un
      }
      where
        stop = do
          dropTemporaryVit cd 3
          modifyAngelFire cd (subtract 1)
        start = do
          modifyAngelFire cd (+1)
          addTemporaryVit cd 3
reevatein = defaultSkill
  { skName = "災厄の剣"
  , skCosts = [Fire, Wind, Sword]
  , skEffect = Just $ \inv -> do
      attackSF (opp inv) 9 inv $ \_ -> do
        mtgt <- chooseTarget inv
        whenJust mtgt $ \tgt -> case tgt of
          PanelTarget cd1 m n -> losePanel cd1 m n >>= performDelays
          PlanTarget cd1 n -> losePlan cd1 n
          SkillTarget cd1 n -> loseSlot cd1 n >>= performDelays
      selfAttack (self inv) 9 inv
      return ()
  , skAttrs = S.fromList [AttackingS]
  , skTargets = Just $ \cd -> do
      let ocd = opponent cd
      ps <- panelTargets ocd ((/=0) . paBound)
      ss <- skillTargets ocd ((/=0) . ssBound)
      pls <- planTargets ocd ((/=0) . psBound)
      return $ S.unions [ps, ss, pls]
  }
flameDevourer = defaultSkill
  { skName = "貪る炎"
  , skCosts = [Fire, Earth, Sword]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 6 inv $ \ndmg -> tryRecover (self inv) ndmg inv
  , skAttrs = S.fromList [AttackingS, RecoveringS]
  }
sharumakihyuNoTousi = defaultSkill
  { skName = "シャルマキヒュの凍視"
  , skCosts = [Water, Wind, Earth]
  , skEffect = Nothing
  , skAttrs = S.fromList [ReactiveS, ObstructiveS]
  , skReactionAgainstSkill = Just $ \_ oinv -> let
        costs = invCosts oinv
        ocd = invControler oinv
      in withCurrentSP ocd False $ \csp -> do
        kinds <- liftM (map paKind . catMaybes) $ mapM (uncurry $ getPanel ocd) costs
        let to_press pa = elem (paKind pa) kinds && not (paUsed pa)
        ps <- liftM (filter to_press) $ getPanels ocd
        mapM_ (makeUsed ocd csp . paIndex) ps
        return False
  }
stormBringer = defaultSkill
  { skName = "嵐の運び手"
  , skCosts = [Water, Wind, Sword]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 5 inv $ \_ -> decrVit (opp inv) 1
  , skAttrs = S.fromList [AttackingS, ObstructiveS]
  }
iwaugachiNoKen = defaultSkill
  { skName = "岩穿ちの剣"
  , skCosts = [Water, Earth, Sword]
  , skEffect = Just $ \inv -> 
      attackEx_ (opp inv) 5 inv defaultAO{ aoBlockable = False }
  , skAttrs = S.fromList [AttackingS]
  }
weatheringBlade = defaultSkill
  { skName = "風蝕刃"
  , skCosts = [Wind, Earth, Sword]
  , skEffect = Just $ \inv ->
      attackSF (opp inv) 3 inv $ \_ -> whenLucky 2 3 $ do
        tgt <- chooseTargetSkill inv good
        whenJust tgt $ \idx -> putWeaken (opp inv) idx
  , skAttrs = S.fromList [AttackingS, MutativeS]
  , skTargets = Just $ \cd -> skillTargets (opponent cd) (good . ssSkill)
  }
  where
    good sk = skIsAttacking sk || skIsCounter sk || skIsDefensive sk || skIsRecovering sk
flashFire = defaultSkill
  { skName = "炎上"
  , skCosts = [Fire, Fire, Fire]
  , skEffect = Just $ \inv -> do
      ds <- to_fire (self inv)
      ds1 <- to_fire (opp inv)
      performDelays $ ds ++ ds1
  , skAttrs = S.fromList [ObstructiveS]
  }
  where
    to_fire cd = withCurrentSP cd [] $ \csp -> do
      ps <- getPanels cd
      let f pa = rollPanelTo cd csp (paIndex pa) Fire
      concatMapM f ps
reflection = defaultSkill
  { skName = "水鏡の盾"
  , skCosts = [Water, Water, Water]
  , skEffect = Nothing
  , skReaction = Just $ \inv ndmg _ -> return $ let
        atk int = attackInt_ (opp inv) int (intInput int - intOutput int) inv
      in (0, False, atk)
  , skAttrs = S.fromList [ReactiveS, DefensiveS, CounterS]
  }
silence = defaultSkill
  { skName = "静謐"
  , skCosts = [Wind, Wind, Wind]
  , skEffect = Nothing
  , skReactionAgainstSkill = Just $ \_ oinv -> do
      putSkillBind (invControler oinv) (invSlot oinv)
      return True
  , skReactionConditionAgainstSkill = return . useIf . skIsMagic . invSkill
  , skAttrs = S.fromList [ReactiveS, ObstructiveS]
  }
kyutosuNoYoumaku = defaultSkill
  { skName = "キュトスの羊膜"
  , skCosts = [Earth, Earth, Earth]
  , skEffect = Nothing
  , skReaction = Just $ \inv ndmg _ -> do
      register inv
      (rdmg, act) <- react ndmg
      return (rdmg, False, act)
  , skReactionCondition = \atk -> do
      have_held <- liftM (any active_kny) $ getHSs (atkTarget atk)
      return $ if have_held then DontUseByDefault else UseByDefault
  , skAttrs = S.fromList [ReactiveS, DefensiveS, ContinuousS]
  }
  where
    register inv = hold (self inv) $ \un -> defaultHS
      { hsSource = inv
      , hsReaction = \n _ -> react n 
      , hsInterfereCondition = const True
      , hsUnique = un
      }
    react ndmg = return (min 1 ndmg, \_ -> return ())
    active_kny hs = invSkill (hsSource hs) == kyutosuNoYoumaku && hsEffective hs
cutthroat = defaultSkill
  { skName = "暗殺"
  , skCosts = [Sword, Sword, Sword]
  , skEffect = Just $ \inv -> 
      attackEx_ (opp inv) 2 inv defaultAO{ aoDamageHook = hook (opp inv) }
  , skAttrs = S.fromList [AttackingS]
  }
  where
    hook ocd n = do
      r <- getRatio ocd
      return $ n * r
    getRatio ocd = do
      panels <- getPanels ocd
      return $ length $ filter paUsed panels

chooseTarget :: Invocation -> Game (Maybe SkillTarget)
chooseTarget inv = do
  tgts <- targets inv
  allt <- fromJust (skTargets $ invSkill inv) $ invControler inv
  let tgts' = S.intersection tgts allt
  chooseMaybe $ S.toList $ if S.null tgts' then allt else tgts'

addTemporaryVit :: CharDescr -> Int -> Game ()
addTemporaryVit cd n = getVit cd >>= setVit cd . (+n)

dropTemporaryVit :: CharDescr -> Int -> Game ()
dropTemporaryVit cd n = do
  vit <- getVit cd
  base <- getBaseVit cd
  setVit cd $ max base (vit - n)

restoreSP :: CharDescr -> Int -> Game ()
restoreSP cd n = restoreG cd n undeleteSP "スキルパターン"

restoreSlot :: CharDescr -> Int -> Game ()
restoreSlot cd n = restoreG cd n undeleteSlot "スキル"

restorePlan :: CharDescr -> Int -> Game ()
restorePlan cd n = restoreG cd n undeletePlan "プラン"

restorePanel :: CharDescr -> Int -> Int -> Game ()
restorePanel cd m n = do
  mesPa cd m n $ "が復活した"
  undeletePanel cd m n

restoreG :: CharDescr -> Int -> (CharDescr -> Int -> Game ()) -> String -> Game ()
restoreG cd n undelete name = do
  mesC cd $ "の" ++ name ++ show n ++ "が復活した"
  undelete cd n

chooseTargetPanel :: Invocation -> (Panel -> Bool) -> Game (Maybe (Int, Int))
chooseTargetPanel inv prop = withCurrentSP (opp inv) Nothing $ \csp -> do
  tgt <- targetPanels inv prop
  ps <- getPanels (opp inv)
  let pps = map paIndex $ filter prop ps
  let cands = if null tgt then pps else tgt
  chooseMaybe $ map ((,) csp) cands

targetPanels :: Invocation -> (Panel -> Bool) -> Game [Int]
targetPanels inv prop = withCurrentSP (opp inv) [] $ \csp -> do
  tgts <- targets inv
  panels <- getPanels (opp inv)
  let
    f pa = prop pa && S.member (PanelTarget (opp inv) csp n) tgts
      where
        n = paIndex pa
  return $ map paIndex $ filter f panels

chooseTargetSkill :: Invocation -> (Skill -> Bool) -> Game (Maybe Int)
chooseTargetSkill inv prop = do
  tgt <- targetSkills inv prop
  cands <- if not $ null tgt then return tgt else do
    liftM (map ssIndex . filter (prop . ssSkill)) $ getSSs $ opp inv
  chooseMaybe cands

targetSkills :: Invocation -> (Skill -> Bool) -> Game [Int]
targetSkills inv prop = do
  tgts <- targets inv
  sss <- getSSs (opp inv)
  let f ss = prop (ssSkill ss) && S.member (SkillTarget (opp inv) $ ssIndex ss) tgts
  return $ map ssIndex $ filter f sss

chooseTargetPlan :: Invocation -> Game (Maybe Int)
chooseTargetPlan inv = do
  tgt <- targetPlans inv
  cands <- if not $ null tgt then return tgt else do
    plans <- getPresentPlans (opp inv)
    return $ map psIndex plans
  chooseMaybe cands

targetPlans :: Invocation -> Game [Int]
targetPlans inv = do
  tgts <- targets inv
  return $ mapMaybe f $ S.toList tgts
  where
    f (PlanTarget cd1 n)
      | cd1 == self inv = Just n
    f _ = Nothing

panelTargets :: CharDescr -> (Panel -> Bool) -> Game (S.Set SkillTarget)
panelTargets cd prop = withCurrentSP cd S.empty $ \csp -> do
  ps <- getPanels cd
  let build n = PanelTarget cd csp n
  return $ S.fromList $ map (build . paIndex) $ filter prop ps

skillTargets :: CharDescr -> (SkillState -> Bool) -> Game (S.Set SkillTarget)
skillTargets cd prop = do
  sss <- getSSs cd
  let build ss = SkillTarget cd (ssIndex ss)
  return $ S.fromList $ map build $ filter prop sss

planTargets :: CharDescr -> (PlanState -> Bool) -> Game (S.Set SkillTarget)
planTargets cd prop = do
  pss <- getPresentPlans cd
  return $ S.fromList $ map (PlanTarget cd . psIndex) $ filter prop pss

whenLucky :: Int -> Int -> Game () -> Game ()
whenLucky m n action = do
  v <- rnd n
  when (v < m) action

block :: Int -> Int -> Int
block d n = max 0 $ n - d

genAttack :: Maybe Interference -> Int -> Invocation -> CharDescr -> Game Attack
genAttack cause n src tgt = do
  (now, turn) <- tick
  let
    atk = Atk
      { atkTime = now
      , atkTurn = turn
      , atkInv = src
      , atkClaimedDmg = n
      , atkCause = cause
      , atkTarget = tgt
      }
  recordEvent atk
  return atk

attack_ :: CharDescr -> Int -> Invocation -> Game ()
attack_ cd n src = attackEx_ cd n src defaultAO

attackInt_ :: CharDescr -> Interference -> Int -> Invocation -> Game ()
attackInt_ cd int n src = attackEx_ cd n src defaultAO{ aoCause = Just int }

attackSF :: CharDescr -> Int -> Invocation -> (Int -> Game ()) -> Game ()
attackSF cd n src side = attackEx_ cd n src defaultAO{ aoSideEffect = side }

attackInt :: CharDescr -> Interference -> Int -> Invocation -> Game () -> (Int -> Game ()) -> Game (Int, Bool)
attackInt cd int n src nega side = attackEx cd n src defaultAO{ aoCause = Just int, aoOnNegation = nega, aoSideEffect = side }

attackEx_ :: CharDescr -> Int -> Invocation -> AttackOptions -> Game ()
attackEx_ cd n src opts = attackEx cd n src opts >> return ()

data AttackOptions = AO
  { aoBlockable :: Bool
  , aoCause :: Maybe Interference
  , aoOnNegation :: Game ()
  , aoSideEffect :: Int -> Game ()
  , aoDamageHook :: Int -> Game Int
  }

defaultAO = AO
  { aoBlockable = True
  , aoCause = Nothing
  , aoOnNegation = return ()
  , aoSideEffect = const $ return ()
  , aoDamageHook = return
  }

attackEx :: CharDescr -> Int -> Invocation -> AttackOptions -> Game (Int, Bool)
attackEx cd n src
  AO{ aoBlockable = blockable, aoCause = cause, aoOnNegation = nega
    , aoSideEffect = side, aoDamageHook = dhook } = do
  nW <- applyWeakenByInv src n
  af <- getAngelFire (opponent cd)
  let nA = nW + if skIsSword $ invSkill src then af * 2 else 0
  atk <- genAttack cause nA src cd
  hes <- heldEffectsReaction cd atk
  old_res <- registeredReactions src
  new_res <- activateReactiveSkills cd atk
  (nR, neg, cas) <- processReactions cd nA atk blockable $ hes ++ old_res ++ new_res
  final_dmg <- dhook nR
  dmg <- if final_dmg == 0 then return Nothing else liftM Just $ damage cd (ByAttack atk) final_dmg
  if neg then nega else side final_dmg
  genAtkResult atk dmg neg
  sequence cas
  return (final_dmg, neg)

genAtkResult :: Attack -> Maybe Damage -> Bool -> Game AttackResult
genAtkResult atk mdmg neg = do
  recordEvent ar
  return ar
  where
    ar = AR{ arAttack = atk, arDamage = mdmg, arNegated = neg }

selfAttack :: CharDescr -> Int -> Invocation -> Game Int
selfAttack cd n src = do
  nW <- applyWeakenByInv src n
  atk <- genAttack Nothing nW src cd
  res <- heldEffectsReaction cd atk
  (nH, False, cas) <- processReactions cd nW atk True res
  dmg <- if nH == 0 then return Nothing else liftM Just $ damage cd (ByAttack atk) nH
  genAtkResult atk dmg False
  sequence cas
  return nH

tryRecover :: CharDescr -> Int -> Invocation -> Game ()
tryRecover cd n inv = do
  nW <- applyWeakenByInv inv n
  recover cd nW

recover :: CharDescr -> Int -> Game ()
recover cd n = do
  chp <- getHp cd
  maxhp <- liftM vitToMaxHp $ getVit cd
  let newhp = max chp $ min maxhp $ chp + n
  let d = newhp - chp
  when (d > 0) $ do
    mesC cd $ "は" ++ show d ++ "点回復した"
    setHp cd newhp

genInt :: Int -> Int -> Bool -> Invocation -> InterferenceType -> Attack -> Game Interference
genInt inp outp neg inv typ tgt = do
  (now, turn) <- tick
  let
    int = Int
      { intTime = now
      , intTurn = turn
      , intInput = inp
      , intOutput = outp
      , intNegated = neg
      , intInv = inv
      , intType = typ
      , intTarget = tgt
      }
  recordEvent int
  return int

type ReactionInfo = (Invocation, InterferenceType, AttackReaction)

registeredReactions :: Invocation -> Game [ReactionInfo]
registeredReactions inv = do
  rs <- getReactions inv
  return $ map f rs
  where
    f (re, inv) = (inv, InvokedI, re)

heldEffectsReaction :: CharDescr -> Attack -> Game [ReactionInfo]
heldEffectsReaction cd atk = do
  hs <- getHSs cd
  liftM catMaybes $ mapM f $ sortOn hsUnique hs
  where
    f h
      | hsEffective h && hsInterfereCondition h atk = 
          return $ Just (hsSource h, HeldI, hsToAttackReaction h)
      | otherwise = return Nothing
    hsToAttackReaction hs inv ndmg atk = do
      (n, act) <- hsReaction hs ndmg atk
      return (n, False, act)

activateReactiveSkills :: CharDescr -> Attack -> Game [ReactionInfo]
activateReactiveSkills cd atk = do
  sds <- plannedSkills cd (isJust . skReaction) Nothing (Just atk)
  liftM catMaybes $ mapM f sds
  where
    f (sd@(SD _ _ sk), forcep) = 
      guarded (isJust $ skReaction sk) $ do
        ready <- sdReady sd
        guarded ready $ do
          cond <- skReactionCondition sk atk
          guarded (evalSkillCondition forcep cond) $ do
            (re, inv) <- reactiveInvoke sd atk
            return $ Just (inv, InvokedI, re)

    guarded c u = if c then u else return Nothing

processReactions :: CharDescr -> Int -> Attack -> Bool -> [ReactionInfo] -> Game (Int, Bool, [Game ()])
processReactions cd ndmg atk blockable ris = foldM f (ndmg, False, []) ris
  where
    f (n, neg, as) (inv, itype, re) = do
      (n', neg', a) <- re inv n atk 
      block <- applyWeakenByInv inv (n - n')
      let out = if blockable then n - block else n
      let real_neg = neg' && not neg
      when real_neg $ mesC (opponent cd) $ "の発動した" ++ show sskill ++ "の追加効果は無効になった"
      int <- genInt n out real_neg inv itype atk
      return (out, neg || neg', as ++ [a int])
    sskill = invSkill $ atkInv atk

makeUsed :: CharDescr -> Int -> Int -> Game ()
makeUsed cd m n = do
  mesPa cd m n $ "は使用済みになった"
  toUsed cd m n

putWeaken :: CharDescr -> Int -> Game ()
putWeaken cd idx = do
  mesC cd $ "のスキル" ++ show idx ++ "はweakenを得た"
  addSkillWeaken cd idx 1

putSkillBind :: CharDescr -> Int -> Game ()
putSkillBind cd idx = do
  mesC cd $ "のスキル" ++ show idx ++ "はBindを得た"
  addSkillBind cd idx 1

putPlanBind :: CharDescr -> Int -> Game ()
putPlanBind cd idx = do
  mesC cd $ "のプラン" ++ show idx ++ "はBindを得た"
  addPlanBind cd idx 1

putPoison :: CharDescr -> Int -> Int -> Game ()
putPoison cd m n = do
  mesPa cd m n $ "はPoisonを得た"
  modifyPanel cd m n [AddPoison 1]

putConfuse :: CharDescr -> Int -> Int -> Game ()
putConfuse cd m n = do
  mesPa cd m n $ "はConfuseを得た"
  modifyPanel cd m n [AddConfuse 1]

putPanelBind :: CharDescr -> Int -> Int -> Game ()
putPanelBind cd m n = do
  mesPa cd m n $ "はBindを得た"
  modifyPanel cd m n [AddBind 1]

targets :: Invocation -> Game (S.Set SkillTarget)
targets = askTargets

self :: Invocation -> CharDescr
self = invControler

opp :: Invocation -> CharDescr
opp = opponent . self

whenJust :: (Monad m) => Maybe a -> (a -> m ()) -> m ()
whenJust x f = maybe (return ()) f x

defaultHS :: HeldSkill
defaultHS = HS
  { hsSource = error "source defaultHS"
  , hsStart = return ()
  , hsStop = return ()
  , hsFinish = return ()
  , hsReaction = error "reaction defaultHS"
  , hsInterfereCondition = const False
  , hsEffective = True
  , hsUnique = error "unique defaultHS"
  }

unregister :: CharDescr -> Unique -> Game ()
unregister cd un = unholdByUnique cd DropByNegation un >>= performDelays

applyWeaken :: CharDescr -> Int -> Int -> Game Int
applyWeaken cd idx n = do
  w <- liftM (maybe 0 ssWeakened) $ getSS cd idx
  return $ max 0 $ n - w

applyWeakenByInv :: Invocation -> Int -> Game Int
applyWeakenByInv inv n = applyWeaken (invControler inv) (invSlot inv) n

atkIsSword, atkIsFire :: Attack -> Bool
atkIsSword = skIsSword . invSkill . atkInv
atkIsFire = skIsFire . invSkill . atkInv

useIf :: Bool -> SkillCondition
useIf False = NeverUse
useIf True = UseByDefault
