import Plan
import Compiler
import Game
import Parser
import Skills
import Match
import AdvancedMatch
import GameUtils
import ExprUtils
import CharBuilder

import System.IO.SSC.Haskell98 as SSC
import System.IO as IO
import Data.Monoid(Monoid(..))
import Control.Monad
import Control.Monad.Reader
import System.Random

simpleCharacter :: String -> (Int, Int, Int) -> [SkillPattern] -> [Skill] -> [String] -> Character
simpleCharacter name params sps ss pss = buildCharacter (Just name) params (map ((,) Nothing) sps) ss (map mkPlan pss)

mkPlan :: String -> Plan
mkPlan s = Plan $ exprCode $ parseExp s

shellon :: Character
shellon = simpleCharacter "暗殺者シェロン" (3, 3, 3) 
  [[Earth, Water, Sword], [Earth, Earth, Water], [Sword, Sword, Sword]]
  [poisonousFang, restore, rest, avenger, slash]
  [ "not any (is_attack &&& not . is_sword), skills opp => switch_to 1"
  , "HP self; <= 3 => dont_use 毒牙"
  , "not has_skill opp, 沙羅双樹;; && HP opp; <= 6 && 1 <= count (eq 地), current_panel_kinds opp;; && first => switch_to 2"
  ]

daemonrose :: Character
daemonrose = simpleCharacter "デーモンローズ" (5, 2, 2)
  [[Earth, Earth], [Sword, Sword]]
  [drainLife, avenger, slash, sarasouju]
  [ "delta = \\t ch -> (beginning t; # HP ch) - (end t; # HP ch) | " ++
      "worse = \\t -> delta t, self; > delta t, opp | " ++
      "(worse prev current_turn) && worse prev prev current_turn => switch_to 1"
  , "any (is_sword . attack_skill), filter (not . eq nil), image damage_attack, turn_damages self, prev current_turn => prioritize 沙羅双樹"
  ]

getInfo1 :: IO GameInfo
getInfo1 = getInfo defaultOption{ optGen = Just gen } shellon daemonrose stop
  where
    stop = do
      now <- getNextTime
      hss <- getHSs Second
      return $ now > toEnum 40 && not (null hss)

main = shell

gen :: RndGen
gen = read "white"

test :: GameInfo -> String -> IO ()
test gi s = do
  runGame game gi
  where
    game = case parseExpE s of
      Left msg -> puts $ "parse error: " ++ msg
      Right exp -> do
        let (msgs, cc) = compile exp
        mapM_ puts msgs
        r1 <- runEval First cc
        case r1 of
          Left x -> puts $ "error: " ++ x
          Right v -> puts $ show v
    puts = liftIO . SSC.putStrLn

shell = do
  IO.hSetBuffering IO.stdin IO.LineBuffering
  gi <- getInfo1
  let
    loop = do
      SSC.putStr "> "
      SSC.hFlush SSC.stdout
      ln <- SSC.getLine
      when (not $ null ln) $ test gi ln
      loop
  loop
